# -*- coding: utf8 -*-

import time
from functools import wraps

from noseapp.utils.common import TimeoutException
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.remote.webelement import WebElement


def make_object(web_element):
    """
    Convert web element to object.

    Example:

        input = driver.find_element...
        value = make_object(input).value
        css_value = make_object(input).css.background_image

    :type web_element: selenium.webdriver.remote.webdriver.WebElement
    """
    # TODO: __setattr__
    class WebElementToObject(object):

        def __init__(self, web_element):
            self.__web_element = web_element

        @property
        def css(self):
            return WebElementCssToObject(self.__web_element)

        def __getattr__(self, item):
            atr = self.__web_element.get_attribute(item)

            if atr:
                return atr

            raise AttributeError('{} "{}"'.format(repr(self.__web_element), item))

    class WebElementCssToObject(object):

        def __init__(self, web_element):
            self.__web_element = web_element

        def __getattr__(self, item):
            return self.__web_element.value_of_css_property(
                item.replace('_', '-'),
            )

    return WebElementToObject(web_element)


def get_driver_from_web_element(web_element):
    """
    :type web_element: selenium.webdriver.remote.webdriver.WebElement
    """
    if hasattr(web_element, 'orig'):
        web_element = web_element.orig()

    if isinstance(web_element, WebElement):
        driver = web_element._parent
    else:
        driver = web_element

    if isinstance(driver, WebDriver):
        return driver

    return get_driver_from_web_element(driver)


def get_driver_from_query(query):
    """
    :param query: instance of QueryProcessor
    """
    return get_driver_from_web_element(query.client)


def get_config(client):
    """
    Get config from WebDriver instance

    :param client: WebDriver or WebElement
    """
    driver = get_driver_from_web_element(client)

    return driver.config


def polling(callback=None, timeout=30, sleep=0.01):
    """
    Do sleep while wrapped function will be raised
    exception of WebDriverException class.

    Use timeout param for setting max seconds to waiting.
    This function will be used like decorator if callback is None.
    """
    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            t_start = time.time()

            while time.time() <= t_start + timeout:
                try:
                    return f(*args, **kwargs)
                except WebDriverException:
                    time.sleep(sleep)
                    continue
            else:
                raise

        return wrapped

    if callback is not None:
        return wrapper(callback)

    return wrapper


class ReRaiseWebDriverException(BaseException):
    pass


def re_raise_wd_exc(callback=None, exc_cls=ReRaiseWebDriverException, message=None):
    """
    Except WebDriverException and re raising to custom exception.
    Also, TimeoutException will be excepting for re raise polling.
    """
    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except (TimeoutException, WebDriverException) as e:
                raise exc_cls(
                    u"""
    Re raise web driver exception:

    * Message: {}
    * Original exception class: {}
    * Original message: {}
                    """.format(
                        message or '',
                        e.__class__.__name__,
                        e.message,
                    ),
                )

        return wrapped

    if callable(callback):
        return wrapper(callback)

    return wrapper



# -*- coding: utf-8 -*-

from copy import deepcopy
from contextlib import contextmanager

from noseapp_selenium import QueryProcessor
from noseapp_selenium.proxy import to_proxy_object
from noseapp_selenium.forms.fields import FormField
from noseapp_selenium.forms.iterators import FieldsIterator
from noseapp_selenium.forms.fields import SimpleFieldInterface
from noseapp_selenium.forms.iterators import RequiredFieldsIterator
from noseapp_selenium.forms.iterators import FieldsWithContainsInvalidValueIterator


def make_field(group_class, weight=None, name=None):
    """
    Usage group like field

    :type form_class: UIForm
    """
    if not issubclass(group_class, FieldsGroup):
        raise ValueError('group_class is not FieldsGroup subclass')

    return GroupContainer(group_class, weight, name)


@contextmanager
def preserve_original(group, ignore_exc=None):
    """
    Save original group after changes inside contextmanager

    :type group: form.UIForm
    :type ignore_exc: BaseException class or tuple
    """
    try:
        if ignore_exc:
            try:
                yield
            except ignore_exc:
                pass
        else:
            yield
    finally:
        group.reload()


def iter_fields(group, exclude=None):
    """
    Example:

        form = MyForm(driver)
        for field in iter_fields(form):
            field.fill()
            ...

    :type exclude: list or tuple
    :return: noseapp_selenium.forms.iterators.FieldsIterator
    """
    assert isinstance(group, FieldsGroup)
    return FieldsIterator(group, exclude=exclude)


def iter_required(group, exclude=None):
    """
    Example:

        form = MyForm(driver)
        for field in iter_required(form):
            form.fill(exclude=(field,))
            form.submit()
            ...

    :type exclude: list or tuple
    :return: noseapp_selenium.forms.iterators.RequiredFieldsIterator
    """
    assert isinstance(group, FieldsGroup)
    return RequiredFieldsIterator(group, exclude=exclude)


def iter_invalid(form, exclude=None):
    """
    Example:

        form = MyForm(driver)
        for field in iter_invalid(form):
            form.fill(exclude=(field,))
            field.fill(field.invalid_value)
            form.submit()
            ...

    :type exclude: list or tuple
    :return: noseapp_selenium.forms.iterators.FieldsWithContainsInvalidValueIterator
    """
    assert isinstance(form, FieldsGroup)
    return FieldsWithContainsInvalidValueIterator(form, exclude=exclude)


class GroupContainer(object):
    """
    Container for group class. Major task is lazy instantiation object.
    """

    def __init__(self, form_class, weight, name):
        self.__form_class = form_class
        self.__weight = weight
        self.__name = name

    @property
    def name(self):
        return self.__name

    def __call__(self, driver, parent):
        return self.__form_class(
            driver,
            parent=parent,
            name=self.__name,
            weight=self.__weight,
        )


class GroupMemento(dict):
    """
    Saved state for the fields group
    """

    def _set_field(self, filed):
        """
        :type filed: fields.BaseField
        """
        self[filed] = {
            'value': filed.value,
            'required': filed.required,
            'error_mess': filed.error_mess,
            'invalid_value': filed.invalid_value,
        }

    def get_field(self, filed):
        return self.get(filed)

    def add_field(self, field):
        assert isinstance(field, FormField)
        self._set_field(field)

    def restore(self, field_list):
        for field in field_list:
            orig = self.get_field(field)

            if orig:
                field.value = orig['value']
                field.required = orig['required']
                field.error_mess = orig['error_mess']
                field.invalid_value = orig['invalid_value']
            else:
                continue


class GroupObserver(object):
    """
    Publish method call for field or field group through handler at this class
    """

    def __init__(self, group):
        assert isinstance(group, FieldsGroup)

        self._group = group

        # listeners
        self._parents = []
        self._children = []

    def add_child(self, form):
        assert isinstance(form, FieldsGroup)
        self._children.append(form)

    def add_parent(self, form):
        assert isinstance(form, FieldsGroup)
        self._parents.append(form)

    def reload_handler(self):
        for child in self._children:
            child.reload()

    def reset_memo_handler(self):
        for child in self._children:
            child.reset_memo()

    def fill_handler(self):
        if self._group._settings.get('remember', True):
            for parent in self._parents:
                parent._fill_memo.add(self._group)

    def clear_handler(self):
        for parent in self._parents:
            try:
                parent._fill_memo.remove(self._group)
            except KeyError:
                pass

    def fill_field_handler(self, field):
        assert isinstance(field, FormField)

        if self._group._settings.get('remember', True):
            self._group._fill_memo.add(field)

    def clear_field_handler(self, field):
        assert isinstance(field, FormField)

        try:
            self._group._fill_memo.remove(field)
        except KeyError:
            pass


class FieldsGroup(SimpleFieldInterface):
    """
    Merging fields to group and simple field
    interface providing for each field through this object.
    """

    def __init__(self, driver, weight=None, name=None, parent=None):
        self._fields = []

        self._driver = to_proxy_object(driver)
        self._weight = weight
        self._fill_memo = set()
        self._memento = GroupMemento()
        self._observer = GroupObserver(self)

        self.__query = QueryProcessor(driver)

        if parent is not None:
            self._observer.add_parent(parent)

        if not hasattr(self, 'name'):
            self.name = name

        meta = getattr(self, 'Meta', object())

        self._settings = {
            'remember': getattr(meta, 'remember', True),
            'allow_raises': getattr(meta, 'allow_raises', True),
            'wrapper': getattr(meta, 'wrapper', None),
            'exclude': getattr(meta, 'exclude', tuple()),
        }

        exclude_atr = (
            'query',
        )

        for atr in (a for a in dir(self) if not a.startswith('_') and a not in exclude_atr):
            maybe_field = getattr(self, atr, None)

            if isinstance(maybe_field, FormField) or isinstance(maybe_field, GroupContainer):
                if maybe_field.name in self._settings['exclude']:
                    try:
                        delattr(self, atr)
                    except AttributeError:
                        pass
                    continue

                self.add_field(atr, deepcopy(maybe_field))

    def add_field(self, name, field):
        if isinstance(field, FormField):
            setattr(self, name, field)
            field.bind(self)
            self._memento.add_field(field)
        elif isinstance(field, GroupContainer):
            field = field(self._driver, self)
            setattr(self, name, field)
            self._observer.add_child(field)
        else:
            raise TypeError('Unknown field type')

        self._fields.append(field)
        self._fields.sort(key=lambda f: f.weight)

    @property
    def weight(self):
        return self._weight

    @property
    def query(self):
        wrapper = self.get_wrapper_element()

        if wrapper:
            return self.__query(wrapper)
        return self.__query

    def submit(self):
        """
        Your submit script here
        """
        pass

    def reload(self):
        self._memento.restore(self._fields)
        self._observer.reload_handler()

    def reset_memo(self):
        self._fill_memo.clear()
        self._observer.reset_memo_handler()

    def update(self, **kwargs):
        """
        Update fields values
        """
        for field_name, value in kwargs.items():
            field = getattr(self, field_name, None)

            if isinstance(field, FormField):
                field.value = value
            elif isinstance(field, FieldsGroup) and isinstance(value, dict):
                field.update(**value)

    def fill(self, exclude=None):
        exclude = exclude or tuple()

        for field in self._fields:
            if (field not in self._fill_memo) and (field not in exclude):
                field.fill()

        self.reset_memo()
        self._observer.fill_handler()

    def clear(self):
        for field in self._fields:
            field.clear()

        self.reset_memo()
        self._observer.clear_handler()

    def get_wrapper_element(self):
        """
        Get web element from meta wrapper
        """
        wrapper = self._settings['wrapper']
        if wrapper:
            return self.__query.from_object(wrapper).first()
        return None



# -*- coding: utf-8 -*-

from noseapp_selenium import drivers
from noseapp_selenium.base import SeleniumEx
from noseapp_selenium.query import QueryProcessor
from noseapp_selenium.page_object import PageObject
from noseapp_selenium.page_object import PageRouter


__all__ = (
    SeleniumEx,
    PageObject,
    PageRouter,
    QueryProcessor,
)


_BASE_CONFIG = {
    'REMOTE_WEBDRIVER': {
        'options': {},
        'capabilities': {
            drivers.IE: {},
            drivers.OPERA: {},
            drivers.CHROME: {},
            drivers.FIREFOX: {},
            drivers.PHANTOMJS: {},
        },
    },

    'IE_WEBDRIVER': {},

    'CHROME_WEBDRIVER': {},

    'FIREFOX_WEBDRIVER': {},

    'PHANTOMJS_WEBDRIVER': {},

    'OPERA_WEBDRIVER': {},

    'OPTIONS': {},
}


def make_config():
    """
    Create base config
    """
    return Config(_BASE_CONFIG)


class Config(dict):

    def configure(self, **options):
        """
        Extension settings configure
        """
        self['OPTIONS'].update(options)

    def remote_configure(self, options=None, capabilities=None):
        """
        :param options: kwargs for method of WebDriver class
        :param capabilities: update base capabilities
        """
        self['REMOTE_WEBDRIVER']['options'].update(options or {})
        self['REMOTE_WEBDRIVER']['capabilities'].update(capabilities or {})

    def ie_configure(self, **options):
        """
        :param options: kwargs for method of WebDriver class
        """
        self['IE_WEBDRIVER'].update(options)

    def chrome_configure(self, **options):
        """
        :param options: kwargs for method of WebDriver class
        """
        self['CHROME_WEBDRIVER'].update(options)

    def firefox_configure(self, **options):
        """
        :param options: kwargs for method of WebDriver class
        """
        self['FIREFOX_WEBDRIVER'].update(options)

    def phantomjs_configure(self, **options):
        """
        :param options: kwargs for method of WebDriver class
        """
        self['PHANTOMJS_WEBDRIVER'].update(options)

    def opera_configure(self, **options):
        """
        :param options: kwargs for method of WebDriver class
        """
        self['OPERA_WEBDRIVER'].update(options)



# -*- coding: utf-8 -*-

from functools import wraps

from selenium.common.exceptions import NoSuchElementException

from noseapp_selenium.tools import make_object
from noseapp_selenium.query import QueryObject
from noseapp_selenium.query import QueryProcessor


def selector(**kwargs):
    """
    proxy for tag attributes
    """
    return kwargs


def fill_field_handler(f):
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        result = f(self, *args, **kwargs)

        if self._settings.get('remenber', True):
            self._observer.fill_field_handler(self)

        return result
    return wrapper


def clear_field_handler(f):
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        result = f(self, *args, **kwargs)

        self._observer.clear_field_handler(self)

        return result
    return wrapper


class FieldError(BaseException):
    pass


class SimpleFieldInterface(object):

    @property
    def weight(self):
        raise NotImplementedError

    def fill(self, value=None):
        raise NotImplementedError

    def clear(self):
        raise NotImplementedError


class FormField(object):
    """
    Base class for all fields
    """

    class Meta:
        tag = None

    def __init__(self,
                 name,
                 value=None,
                 required=False,
                 selector=None,
                 error_mess=None,
                 invalid_value=None,
                 weight=None):

        self.name = name

        if not isinstance(selector, dict):
            raise ValueError('incorrect selector')

        self.__is_bind = False

        self._query = None
        self._settings = {}
        self._observer = None
        self._weight = weight
        self._selector = selector

        self.value = value
        self.required = required
        self.error_mess = error_mess
        self.invalid_value = invalid_value

    def bind(self, group):
        try:
            self._query = group._driver.query
        except AttributeError:
            self._query = QueryProcessor(group._driver)

        self._observer = group._observer
        self._settings = group._settings

        if callable(self.value):
            self.value = self.value()

        if callable(self.invalid_value):
            self.invalid_value = self.invalid_value()

        self.__is_bind = True

    @property
    def weight(self):
        return self._weight

    def get_web_element(self):
        if not self.__is_bind:
            raise FieldError('Field is not binding to group')

        wrapper = self._settings.get('wrapper', None)

        if isinstance(wrapper, QueryObject):
            query = self._query(
                self._query.from_object(wrapper).first(),
            )
        else:
            query = self._query

        return query.from_object(
            QueryObject(self.Meta.tag, **self._selector),
        ).first()

    @property
    def obj(self):
        return make_object(self.get_web_element())


def field_on_base(*interfaces):
    """
    Create parent class on base interfaces
    """
    if not hasattr(field_on_base, '__classes__'):
        field_on_base.__classes__ = {}

    bases = tuple([FormField] + sorted(list(interfaces)))

    try:
        return field_on_base.__classes__[bases]
    except KeyError:
        field_on_base.__classes__[bases] = type('BaseField', bases, {})
        return field_on_base.__classes__[bases]


class Input(field_on_base(SimpleFieldInterface)):

    class Meta:
        tag = 'input'

    @fill_field_handler
    def fill(self, value=None):
        if value is None:
            value = self.value

        self.get_web_element().send_keys(*value)

    @clear_field_handler
    def clear(self):
        self.get_web_element().clear()


class TextArea(Input):

    class Meta:
        tag = 'textarea'


class Checkbox(field_on_base(SimpleFieldInterface)):

    class Meta:
        tag = 'input'

    @fill_field_handler
    def fill(self, value=None):
        if value is None:
            value = self.value

        el = self.get_web_element()
        current_value = el.is_selected()

        if self._settings.get('allow_raises', True):

            if current_value and value:
                raise FieldError('Oops, checkbox did selected')
            elif not current_value and not value:
                raise FieldError('Oops, checkbox did unselected')

        if (value and not current_value) or (not value and current_value):
            el.click()

    @clear_field_handler
    def clear(self):
        el = self.get_web_element()

        if el.is_selected():
            el.click()


class RadioButton(Checkbox):

    class Meta:
        tag = 'input'

    @fill_field_handler
    def fill(self, value=None):
        if value is None:
            value = self.value

        el = self.get_web_element()
        current_value = el.is_selected()
        changed = False

        if (value and not current_value) or (not value and current_value):
            el.click()
            changed = True

        return changed

    @clear_field_handler
    def clear(self):
        pass


class Select(field_on_base(SimpleFieldInterface)):

    class Meta:
        tag = 'select'

    @fill_field_handler
    def fill(self, value=None):
        if value is None:
            value = self.value

        select = self.get_web_element()

        options = filter(
            lambda opt: opt.get_attribute('value') == value,
            select.find_elements_by_tag_name('option'),
        )

        if options:
            options[0].click()
        else:
            raise NoSuchElementException(
                u'Cant do selected value for field "{}", selector "{}", option "{}"'.format(
                    self.name, str(self._selector), value
                ),
            )

    @clear_field_handler
    def clear(self):
        pass



# -*- coding: utf-8 -*-

import time
from Queue import Queue

from noseapp.utils.common import waiting_for
from noseapp.utils.common import TimeoutException

from noseapp_selenium import QueryProcessor
from noseapp_selenium.query import QueryObject
from noseapp_selenium.proxy import to_proxy_object


def page_element(query_object):
    """
    Factory for creating query
    result from query object instance
    """
    def wrapper(self):
        return self.query.from_object(query_object).first()
    return property(wrapper)


class WaitConfig(object):
    """
    Configuration for wait complete method
    """

    def __init__(self,
                 objects=None,
                 one_of_many=False,
                 timeout=30,
                 ready_state_complete=True):
        self._objects = objects or tuple()
        self._one_of_many = one_of_many
        self._timeout = timeout
        self._ready_state_complete = ready_state_complete

    @property
    def objects(self):
        return self._objects

    @property
    def timeout(self):
        return self._timeout

    @property
    def one_of_many(self):
        return self._one_of_many

    @property
    def ready_state_complete(self):
        return self._ready_state_complete


class PageObjectMeta(type):
    """
    Factory for creating page object class
    """

    def __new__(cls, name, bases, dct):
        new_cls = type.__new__(cls, name, bases, dct)

        cls_items = (
            (a, getattr(new_cls, a, None))
            for a in dir(new_cls)
            if not a.startswith('_')
        )

        for atr, value in cls_items:
            if isinstance(value, QueryObject):
                setattr(new_cls, atr, page_element(value))

        return new_cls


class PageObject(object):
    """
    Base page class
    """

    __metaclass__ = PageObjectMeta

    def __init__(self, driver):
        self._driver = to_proxy_object(driver)
        self.__query = QueryProcessor(driver)

        meta = getattr(self, 'Meta', object())

        self._wrapper = getattr(meta, 'wrapper', None)

        if not hasattr(self, 'wait_complete'):
            wait_config = getattr(meta, 'wait_config', WaitConfig())
            self.wait_complete = WaitComplete(
                self._driver,
                self.__query,
                self.__class__.__name__,
                wait_config,
            )

    @property
    def query(self):
        if isinstance(self._wrapper, QueryObject):
            return self.__query(
                self.__query.from_object(self._wrapper).first(),
            )
        return self.__query


class WaitComplete(object):
    """
    Waiting for load page
    """

    def __init__(self, driver, query, page_name, config):
        self.config = config

        self.__query = query
        self.__driver = driver
        self.__page_name = page_name

    def __call__(self):
        if self.config.ready_state_complete:
            self.__ready_state_complete()

        wait_funcs = {
            False: self.__all,
            True: self.__one_of_many,
        }
        wait_funcs[bool(self.config.one_of_many)]()

    def __repr__(self):
        return '<WaitComplete of <{}>>'.format(self.__page_name)

    def __ready_state_complete(self):
        waiting_for(
            lambda: self.__driver.execute_script(
                'return document.readyState == "complete"',
            ),
            timeout=self.config.timeout,
        )

    def __all(self):
        if not self.config.objects:
            return

        queue = Queue()
        map(queue.put_nowait, self.config.objects)
        t_start = time.time()

        while (time.time() <= t_start + self.config.timeout) or (not queue.empty()):
            obj = queue.get()

            if not self.__query.from_object(obj).exist:
                queue.put(obj)
        else:
            if not queue.empty():
                raise TimeoutException(
                    'Could not wait ready page "{}". Timeout "{}" exceeded.'.format(
                        self.__page_name, self.config.timeout,
                    ),
                )

    def __one_of_many(self):
        if not self.config.objects:
            return

        queue = Queue()
        map(queue.put_nowait, self.config.objects)
        t_start = time.time()

        while time.time() <= t_start + self.config.timeout:
            obj = queue.get()

            if self.__query.from_object(obj).exist:
                break

            queue.put(obj)
        else:
            raise TimeoutException(
                'Could not wait ready page "{}". Timeout "{}" exceeded.'.format(
                    self.__page_name, self.config.timeout,
                ),
            )



# -*- coding: utf-8 -*-

import logging

from noseapp.utils.common import waiting_for
from noseapp.utils.common import TimeoutException
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.remote.webelement import WebElement

from noseapp_selenium.proxy import ProxyObject
from noseapp_selenium.proxy import to_proxy_object
from noseapp_selenium.tools import get_driver_from_web_element


logger = logging.getLogger(__name__)


REPLACE_ATTRIBUTES = {
    '_id': 'id',
    '_class': 'class',
    '_type': 'type',
}

REPLACE_TAGS = {
    'link': 'a',
}

DEFAULT_SLEEP = 0.01
DEFAULT_WAIT_TIMEOUT = 30


class QueryError(BaseException):
    pass


def _error_handler(e, client, css):
    """
    Extend error message
    """
    prefix = u' ' if e.message else u''

    if isinstance(client.orig(), WebElement):
        e.message += u'{}QueryProcessor(From: {}, CSS: {})\n\n--\nSEARCH AREA: {}\n--\n'.format(
            prefix,
            repr(client),
            css,
            client.get_attribute('innerHTML'),
        )
    else:
        e.message += u'{}QueryProcessor(From: {}, CSS: {})'.format(
            prefix,
            repr(client),
            css,
        )


def _execute(client, css, get_all=False, allow_polling=True):
    """
    Execute css query
    """
    logger.debug(u'CSS: {} Get all: {}'.format(css, 'Yes' if get_all else 'No'))

    css_executor = lambda cli: {
        True: cli.find_elements_by_css_selector,
        False: cli.find_element_by_css_selector,
    }

    try:
        if allow_polling:
            result = css_executor(client)[get_all](css)
        elif isinstance(client, ProxyObject):
            with client.disable_polling():
                result = css_executor(client)[get_all](css)
        else:
            result = css_executor(client)[get_all](css)

        return result

    except WebDriverException as e:
        _error_handler(e, client, css)
        raise
    except KeyError:
        raise QueryError('get_all param must be bool type only')


def _replace_attribute(atr_name):
    """
    Replacing attribute name for
    excluding conflict with names of globals
    """
    return REPLACE_ATTRIBUTES.get(atr_name, atr_name).replace('_', '-')


def _replace_tag(tag_name):
    """
    Replacing tag name for usability
    """
    return REPLACE_TAGS.get(tag_name, tag_name)


def _handler(client, tag):
    """
    :type client: selenium.webdriver.remote.webdriver.WebDriver
    :param tag: html tag name
    """
    def handle(**selector):
        query = [_replace_tag(tag)]

        def get_format(value):
            if isinstance(value, _Contains):
                return u'[{}*="{}"]'
            return u'[{}="{}"]'

        query.extend(
            (
                get_format(val).format(_replace_attribute(atr), val)
                for atr, val in selector.items()
            ),
        )

        return QueryResult(client, ''.join(query))

    return handle


class QueryObject(object):
    """
    Structure of css query.

    Use instance of this class with
    from_object method of QueryProcessor.
    """

    def __init__(self, tag, **selector):
        self.tag = tag
        self.selector = selector

    def __unicode__(self):
        return u'<{} {}>'.format(
            self.tag,
            u' '.join(
                [u'{}="{}"'.format(k, v) for k, v in self.selector.items()]
            ),
        )


class _Contains(object):
    """
    Marker for use contains inside css query
    """

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return self.value

    def __str__(self):
        return str(self.value)

    def __unicode__(self):
        return unicode(self.value)


contains = _Contains


class QueryResult(object):
    """
    Execute actions by css query and returning result
    """

    def __init__(self, client, css):
        self._client = client
        self._css = css

    def __getattr__(self, item):
        return getattr(QueryProcessor(self.first()), item)

    @property
    def exist(self):
        """
        Check element exist
        """
        driver = get_driver_from_web_element(self._client)
        driver.implicitly_wait(0)

        try:
            el = _execute(self._client, self._css, allow_polling=False)

            if el:
                driver.config.apply_implicitly_wait()
                return True
            driver.config.apply_implicitly_wait()
            return False
        except WebDriverException:
            driver.config.apply_implicitly_wait()
            return False
        except BaseException:
            driver.config.apply_implicitly_wait()
            raise

    @property
    def with_wait(self):
        self.wait()
        return self

    def wait(self, timeout=None, sleep=None):
        """
        Waiting for web element exist
        """
        try:
            return waiting_for(
                lambda: self.exist,
                sleep=sleep or DEFAULT_SLEEP,
                timeout=timeout or DEFAULT_WAIT_TIMEOUT,
            )
        except TimeoutException:
            raise TimeoutException(
                'Could not wait web element with css "{}"'.format(self._css),
            )

    def get(self, index):
        """
        Get web element by index
        """
        try:
            return _execute(self._client, self._css, get_all=True)[index]
        except IndexError as e:
            e.message = 'Result does not have element with index "{}". Query: "{}".'.format(
                index, self._css,
            )
            raise

    def first(self):
        """
        Get first element on page
        """
        return _execute(self._client, self._css)

    def all(self):
        """
        Get all elements of appropriate query
        """
        return _execute(self._client, self._css, get_all=True)


class QueryProcessor(object):
    """
    Example:

        driver = WebDriver()
        query = QueryProcessor(driver)

        search_wrapper = query.div(id='search').first()
        search_field = query(search_wrapper).input(_class='search').first()
        search_field.send_keys(*'Hello World!')
    """

    def __init__(self, client):
        self._client = to_proxy_object(client)

    def __getattr__(self, item):
        return _handler(self._client, item)

    def __call__(self, client):
        return self.__class__(client)

    @property
    def client(self):
        return self._client

    def from_object(self, obj):
        """
        Creating result from object

        :type obj: QueryObject
        """
        if not isinstance(obj, QueryObject):
            raise TypeError('"{}" is not QueryObject instance'.format(type(obj)))

        return self.__getattr__(obj.tag)(**obj.selector)

    def get_text(self):
        """
        Get text from page or web element
        """
        if isinstance(self._client.orig(), WebElement):
            return self._client.text

        return self._client.find_element_by_tag_name('body').text



# -*- coding: utf-8 -*-

from noseapp_selenium.forms import fields
from noseapp_selenium.forms.group import make_field
from noseapp_selenium.forms.group import FieldsGroup
from noseapp_selenium.forms.group import iter_fields
from noseapp_selenium.forms.group import iter_invalid
from noseapp_selenium.forms.group import iter_required
from noseapp_selenium.forms.group import preserve_original


class UIForm(FieldsGroup):
    pass


__all__ = (
    fields,
    UIForm,
    make_field,
    FieldsGroup,
    iter_fields,
    iter_invalid,
    iter_required,
    preserve_original,
)



# -*- coding: utf-8 -*-

from types import MethodType
from contextlib import contextmanager

from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement

from noseapp_selenium.tools import polling
from noseapp_selenium.tools import get_config


def factory(f):
    """
    Factory for create WebElement instance
    """
    def wrapper(*args, **kwargs):
        result = f(*args, **kwargs)

        if isinstance(result, WebElement):
            return ProxyObject(result)

        if isinstance(result, list):
            we_list = []
            for obj in result:
                if isinstance(obj, WebElement):
                    we_list.append(ProxyObject(obj))
                else:
                    return result
            return we_list

        return result

    return wrapper


def to_proxy_object(obj):
    """
    Convert instance to ProxyObject
    """
    if isinstance(obj, (WebElement, WebDriver)):
        return ProxyObject(obj)

    return obj


class ProxyObject(object):
    """
    Proxy for WebElement or WebDriver instance
    """

    def __init__(self, wrapped):
        self.__dict__['polling'] = True
        self.__dict__['wrapped'] = wrapped
        self.__dict__['config'] = get_config(wrapped)

    @contextmanager
    def disable_polling(self):
        try:
            self.__dict__['polling'] = False

            wrapped = self.__dict__['wrapped']

            if hasattr(wrapped, 'disable_polling'):
                with wrapped.disable_polling():
                    yield
                    self.__dict__['polling'] = True
            else:
                yield
                self.__dict__['polling'] = True
        except:
            self.__dict__['polling'] = True
            raise

    def orig(self):
        return self.__dict__['wrapped']

    def __getattr__(self, item):
        wrapped = self.__dict__['wrapped']
        config = self.__dict__['config']

        allow_polling = config.POLLING_TIMEOUT and self.__dict__['polling']

        attr = getattr(wrapped, item)

        if callable(attr) and type(attr) == MethodType:
            if allow_polling:
                return polling(callback=factory(attr), timeout=config.POLLING_TIMEOUT)

            return factory(attr)

        return attr

    def __setattr__(self, key, value):
        return setattr(self.__dict__['wrapped'], key, value)

    def __repr__(self):
        return 'ProxyObject: {}'.format(
            repr(self.__dict__['wrapped']),
        )



# -*- coding: utf-8 -*-

from noseapp_selenium.page_object.base import PageObject
from noseapp_selenium.page_object.base import WaitConfig
from noseapp_selenium.page_object.router import PageRouter


__all__ = (
    PageRouter,
    PageObject,
    WaitConfig,
)



# -*- coding: utf-8 -*-

import logging
from functools import wraps
from urllib2 import URLError

from noseapp.core import ExtensionInstaller
from noseapp.utils.common import waiting_for
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities

from noseapp_selenium import drivers
from noseapp_selenium.query import QueryProcessor
from noseapp_selenium.proxy import to_proxy_object


logger = logging.getLogger(__name__)


_DRIVER_TO_CAPABILITIES = {
    drivers.OPERA: DesiredCapabilities.OPERA,
    drivers.CHROME: DesiredCapabilities.CHROME,
    drivers.FIREFOX: DesiredCapabilities.FIREFOX,
    drivers.PHANTOMJS: DesiredCapabilities.PHANTOMJS,
    drivers.IE: DesiredCapabilities.INTERNETEXPLORER,
}

DEFAULT_IMPLICITLY_WAIT = 30
DEFAULT_MAXIMIZE_WINDOW = True
DEFAULT_POLLING_TIMEOUT = None
DEFAULT_DRIVER = drivers.CHROME

GET_DRIVER_TIMEOUT = 10
GET_DRIVER_SLEEP = 0.5


class SeleniumExError(BaseException):
    pass


def get_capabilities(driver_name):
    try:
        return _DRIVER_TO_CAPABILITIES[driver_name]
    except KeyError:
        raise SeleniumExError(
            'capabilities for driver "{}" is not found'.find(driver_name),
        )


def setup_config(f):
    """
    Setup config to driver and apply settings
    """
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        driver = f(self, *args, **kwargs)

        driver.config = DriverConfig(
            driver,
            implicitly_wait=self._implicitly_wait,
            maximize_window=self._maximize_window,
            polling_timeout=self._polling_timeout,
        )
        driver.config.apply()

        return to_proxy_object(driver)

    return wrapper


class DriverConfig(object):

    def __init__(self, driver,
                 implicitly_wait=DEFAULT_IMPLICITLY_WAIT,
                 maximize_window=DEFAULT_MAXIMIZE_WINDOW,
                 polling_timeout=DEFAULT_POLLING_TIMEOUT):
        self.__driver = driver

        self.IMPLICITLY_WAIT = implicitly_wait
        self.MAXIMIZE_WINDOW = maximize_window
        self.POLLING_TIMEOUT = polling_timeout

    def apply(self):
        self.init_query()

        self.apply_implicitly_wait()
        self.apply_maximize_window()

    def apply_implicitly_wait(self):
        if self.IMPLICITLY_WAIT is not None:
            self.__driver.implicitly_wait(self.IMPLICITLY_WAIT)
        else:
            self.__driver.IMPLICITLY_WAIT = 0

    def apply_maximize_window(self):
        if self.MAXIMIZE_WINDOW:
            self.__driver.maximize_window()

    def init_query(self):
        self.__driver.query = QueryProcessor(self.__driver)


class SeleniumEx(object):
    """
    Initialize selenium extension for noseapp
    """

    name = 'selenium'
    config_key = 'SELENIUM_EX'

    def __init__(
            self,
            config,
            use_remote=False,
            driver_name=DEFAULT_DRIVER,
            maximize_window=DEFAULT_MAXIMIZE_WINDOW,
            implicitly_wait=DEFAULT_IMPLICITLY_WAIT,
            polling_timeout=DEFAULT_POLLING_TIMEOUT):
        self._config = config
        self._use_remote = use_remote
        self._driver_name = driver_name.lower()
        self._maximize_window = maximize_window
        self._implicitly_wait = implicitly_wait
        self._polling_timeout= polling_timeout

        logger.debug(
            'Selenium-EX initialize. Config: {}, Use Remote: {}, Driver name: {}'.format(
                config,
                'Yes' if use_remote else 'No',
                driver_name,
            ),
        )

    @classmethod
    def install(cls, app):
        """
        Automatic installation from app instance

        :param app: instance of application
        """
        config = app.config.get(cls.config_key, {})
        options = config.pop('OPTIONS', {})
        installer = ExtensionInstaller(cls, (config, ), options)
        app.shared_extension(cls=installer)
        return installer

    @property
    def config(self):
        return self._config

    @setup_config
    def remote(self):
        remote_config = self._config.get('REMOTE_WEBDRIVER')

        if not remote_config:
            raise SeleniumExError('remote web driver settings not found')

        logger.debug('Remote config: {}'.format(str(remote_config)))

        options = remote_config.get('options', {})
        capabilities = get_capabilities(self._driver_name)
        capabilities.update(
            remote_config['capabilities'][self._driver_name],
        )

        return drivers.RemoteWebDriver(
            desired_capabilities=capabilities,
            **options
        )

    @setup_config
    def ie(self):
        ie_config = self._config.get('IE_WEBDRIVER')

        if not ie_config:
            raise SeleniumExError('ie web driver settings not found')

        logger.debug('Ie config: {}'.format(str(ie_config)))

        return drivers.IeWebDriver(**ie_config)

    @setup_config
    def chrome(self):
        chrome_config = self._config.get('CHROME_WEBDRIVER')

        if not chrome_config:
            raise SeleniumExError('google chrome web driver settings not found')

        logger.debug('Chrome config: {}'.format(str(chrome_config)))

        return drivers.ChromeWebDriver(**chrome_config)

    @setup_config
    def firefox(self):
        firefox_config = self._config.get('FIREFOX_WEBDRIVER', {})

        logger.debug('Firefox config: {}'.format(str(firefox_config)))

        return drivers.FirefoxWebDriver(**firefox_config)

    @setup_config
    def phantom(self):
        phantom_config = self._config.get('PHANTOMJS_WEBDRIVER')

        if not phantom_config:
            raise SeleniumExError('phantom js web driver settings not found')

        logger.debug('PhantomJS config: {}'.format(str(phantom_config)))

        return drivers.PhantomJSWebDriver(**phantom_config)

    @setup_config
    def opera(self):
        opera_config = self._config.get('OPERA_WEBDRIVER')

        if not opera_config:
            raise SeleniumExError('opera web driver settings not found')

        logger.debug('Opera config: {}'.format(str(opera_config)))

        return drivers.OperaWebDriver(**opera_config)

    def _get_local_driver(self):
        driver = getattr(self, self._driver_name, None)

        if driver:
            return driver()

        raise SeleniumExError(
            'incorrect driver name "{}"'.format(self._driver_name),
        )

    def get_driver(self,
                   driver_name=None,
                   timeout=GET_DRIVER_TIMEOUT,
                   sleep=GET_DRIVER_SLEEP):
        if driver_name is not None:
            self._driver_name = driver_name

        def get_driver(func):
            try:
                return func()
            except URLError:
                return None

        if self._use_remote:
            driver = waiting_for(
                lambda: get_driver(self.remote),
                timeout=timeout,
                sleep=sleep,
            )
        else:
            driver = waiting_for(
                lambda: get_driver(self._get_local_driver),
                timeout=timeout,
                sleep=sleep,
            )

        return driver



# -*- coding: utf-8 -*-

import re

from noseapp.utils.common import waiting_for

from noseapp_selenium.query import QueryResult
from noseapp_selenium.query import QueryObject
from noseapp_selenium.page_object.base import PageObject


class PageIsNotFound(BaseException):
    pass


class PageRouter(object):
    """
    Realization relationships of page object to regexp string.
    """

    __rules = {}

    def __init__(self, driver, base_path=None):
        self._driver = driver
        self._base_path = base_path.rstrip('/')

    @classmethod
    def add_rule(cls, rule, page_cls):
        """
        Add rule for page object class

        :param rule: regexp
        :param page_cls: page object class
        """
        if not issubclass(page_cls, PageObject):
            raise ValueError('page is not PageObject subclass')

        cls.__rules[re.compile(r'^{}$'.format(rule))] = page_cls

    @property
    def base_path(self):
        return self._base_path + '/'

    def get(self, path, wait=True, go_to=True):
        """
        Get page object instance by path.

        Create page instance, go to path,
        call wait complete method of page object.

        :type path: str
        """
        for rule in self.__rules:
            if rule.search(path) is not None:
                page_cls = self.__rules[rule]
                page = page_cls(self._driver)
                break
        else:
            raise PageIsNotFound(path)

        if self._base_path is not None and go_to:
            self.go_to(path)

            if type(wait) is bool and wait:
                page.wait_complete()
            elif isinstance(wait, QueryResult):
                wait.wait()
            elif isinstance(wait, QueryObject):
                page._query.from_object(wait).wait()
            elif callable(wait):
                waiting_for(wait)

        return page

    def get_no_wait(self, path):
        """
        Get method wrapper.

        Get page object instance by rule and
        exclude calling wait complete method of page object.
        """
        return self.get(path, wait=False)

    def get_page(self, path):
        """
        Get method wrapper.

        Get page object instance by rule. Don't go to path.
        """
        return self.get(path, go_to=False)

    def go_to(self, path):
        """
        Simple go to path.
        """
        self._driver.get('{}{}'.format(self._base_path, path))
